"use strict";
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _SingleInstanceTaskScheduler_instances, _SingleInstanceTaskScheduler_task, _SingleInstanceTaskScheduler_options, _SingleInstanceTaskScheduler_context, _SingleInstanceTaskScheduler_nextRunData, _SingleInstanceTaskScheduler_taskRunningPromise, _SingleInstanceTaskScheduler_determineFixedIntervalNextRun, _SingleInstanceTaskScheduler_scheduleWithSuccessResult, _SingleInstanceTaskScheduler_scheduleWithErrorResult, _SingleInstanceTaskScheduler_runTask;
Object.defineProperty(exports, "__esModule", { value: true });
exports.SingleInstanceTaskScheduler = void 0;
// Private classes
class AssertionError extends Error {
    constructor(message) {
        super(message);
        this.name = 'AssertionError';
    }
}
// Public classes
/**
 * A task scheduler which have at most 1 running task at any time.
 *
 * Stability: 2 - Stable.
 */
class SingleInstanceTaskScheduler {
    // Reference https://stackoverflow.com/a/52477831/1131246
    constructor(...values) {
        var _a, _b, _c;
        _SingleInstanceTaskScheduler_instances.add(this);
        _SingleInstanceTaskScheduler_task.set(this, void 0);
        _SingleInstanceTaskScheduler_options.set(this, void 0);
        _SingleInstanceTaskScheduler_context.set(this, void 0);
        _SingleInstanceTaskScheduler_nextRunData.set(this, null);
        _SingleInstanceTaskScheduler_taskRunningPromise.set(this, null);
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_task, values[0], "f");
        const options = (_a = values[1]) !== null && _a !== void 0 ? _a : {};
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_options, {
            onSuccess: (_b = options.onSuccess) !== null && _b !== void 0 ? _b : null,
            onError: (_c = options.onError) !== null && _c !== void 0 ? _c : null
        }, "f");
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_context, values[2], "f");
    }
    get successNextRunOptions() {
        return __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_options, "f").onSuccess;
    }
    set successNextRunOptions(value) {
        __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_options, "f").onSuccess = value;
    }
    get errorNextRunOptions() {
        return __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_options, "f").onError;
    }
    set errorNextRunOptions(value) {
        __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_options, "f").onError = value;
    }
    get context() {
        return __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_context, "f");
    }
    /**
     * Whether a next run is scheduled.
     */
    get scheduled() {
        return (__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f") !== null);
    }
    /**
     * The next run time of the task. `null` indicates that there will be no next run.
     */
    get nextRunTime() {
        var _a, _b;
        return (_b = (_a = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f")) === null || _a === void 0 ? void 0 : _a.startTime) !== null && _b !== void 0 ? _b : null;
    }
    /**
     * Whether the task is currently running.
     */
    get running() {
        return __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_taskRunningPromise, "f") !== null;
    }
    /**
     * Schedule the task to run after a given milliseconds or absolute date time. If
     * the task is already scheduled, it will be re-scheduled with attempt number retaining
     * the previous value.
     *
     * @param startDelayOrTime Start time of the next run. A delay in milliseconds, or
     * an absolute date time.
     */
    schedule(startDelayOrTime) {
        var _a;
        const prevAttemptNumber = (_a = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f")) === null || _a === void 0 ? void 0 : _a.attemptNumber;
        let startTime = startDelayOrTime;
        if (typeof startTime === 'number') {
            startTime = new Date(Date.now() + startTime);
        }
        this.cancelNextRun();
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_nextRunData, {
            startTime,
            timer: setTimeout(__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_instances, "m", _SingleInstanceTaskScheduler_runTask).bind(this), startTime.getTime() - Date.now()),
            attemptNumber: prevAttemptNumber !== null && prevAttemptNumber !== void 0 ? prevAttemptNumber : 1
        }, "f");
    }
    /**
     * Cancel the next scheduled run. Task already running will not be cancelled.
     */
    cancelNextRun() {
        var _a;
        if (((_a = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f")) === null || _a === void 0 ? void 0 : _a.timer) != null) {
            clearTimeout(__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f").timer);
        }
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_nextRunData, null, "f");
    }
    /**
     * Run task immediately without waiting for the task returned value. Previous attempt number
     * is retained. If configured, next run will be scheduled after run completed.
     */
    // We intented to return Promise from non-async function
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    run() {
        var _a;
        const prevAttemptNumber = (_a = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f")) === null || _a === void 0 ? void 0 : _a.attemptNumber;
        this.cancelNextRun();
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_nextRunData, {
            startTime: new Date(),
            timer: null,
            attemptNumber: prevAttemptNumber !== null && prevAttemptNumber !== void 0 ? prevAttemptNumber : 1
        }, "f");
        __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_instances, "m", _SingleInstanceTaskScheduler_runTask).call(this);
        if (!(__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_taskRunningPromise, "f") instanceof Promise)) {
            throw new AssertionError('Expect #taskRunningPromise is a Promise');
        }
        return __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_taskRunningPromise, "f");
    }
}
exports.SingleInstanceTaskScheduler = SingleInstanceTaskScheduler;
_SingleInstanceTaskScheduler_task = new WeakMap(), _SingleInstanceTaskScheduler_options = new WeakMap(), _SingleInstanceTaskScheduler_context = new WeakMap(), _SingleInstanceTaskScheduler_nextRunData = new WeakMap(), _SingleInstanceTaskScheduler_taskRunningPromise = new WeakMap(), _SingleInstanceTaskScheduler_instances = new WeakSet(), _SingleInstanceTaskScheduler_determineFixedIntervalNextRun = function _SingleInstanceTaskScheduler_determineFixedIntervalNextRun(thisRunData, options) {
    var _a;
    let nextRun;
    const onPastTime = (_a = options.onPastTime) !== null && _a !== void 0 ? _a : 'RUN_IMMEDIATELY';
    if (onPastTime === 'RUN_IMMEDIATELY') {
        nextRun = new Date(thisRunData.startTime.getTime() + options.interval);
    }
    else if (onPastTime === 'NEXT_RUN_TIME') {
        if (options.interval <= 1) {
            nextRun = new Date();
        }
        else {
            const thisTimeSlot = thisRunData.startTime.getTime();
            const interval = options.interval;
            const now = Date.now();
            const diff = now - thisTimeSlot;
            const increment = (((diff - (diff % interval)) / interval) + 1) * interval;
            const newTimestampMs = thisTimeSlot + increment;
            if (!(newTimestampMs > now)) {
                throw new AssertionError('Expect newTimestampMs is greater than now');
            }
            nextRun = new Date(newTimestampMs);
        }
    }
    else {
        throw new AssertionError('Not implemented case');
    }
    return nextRun;
}, _SingleInstanceTaskScheduler_scheduleWithSuccessResult = function _SingleInstanceTaskScheduler_scheduleWithSuccessResult(taskReturnValue, startTime, endTime) {
    // Determine next run time
    let nextRun;
    if (__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f") === null) {
        // `cancalNextRun()` is called while task is running
        nextRun = null;
    }
    else {
        const thisRunData = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f");
        const options = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_options, "f").onSuccess;
        if (options === null) {
            nextRun = null;
        }
        else if (typeof options === 'function') {
            nextRun = options(taskReturnValue, { startTime, endTime }, __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_context, "f"));
        }
        else if (options.type === 'FIXED_INTERVAL') {
            nextRun = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_instances, "m", _SingleInstanceTaskScheduler_determineFixedIntervalNextRun).call(this, thisRunData, options);
        }
        else if (options.type === 'RUN_END_TIME') {
            nextRun = new Date(endTime.getTime() + options.delay);
        }
        else {
            throw new AssertionError('Not implemented case');
        }
    }
    // Use next run time to set next run data
    if (nextRun === null) {
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_nextRunData, null, "f");
    }
    else {
        if (typeof nextRun === 'number') {
            nextRun = new Date(Date.now() + nextRun);
        }
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_nextRunData, {
            startTime: nextRun,
            timer: setTimeout(__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_instances, "m", _SingleInstanceTaskScheduler_runTask).bind(this), nextRun.getTime() - Date.now()),
            attemptNumber: 1
        }, "f");
    }
}, _SingleInstanceTaskScheduler_scheduleWithErrorResult = function _SingleInstanceTaskScheduler_scheduleWithErrorResult(caughtValue, startTime, endTime) {
    var _a, _b, _c, _d, _e;
    const thisRunData = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f");
    // Determine next run time
    let nextRun;
    if (thisRunData === null) {
        // `cancalNextRun()` is called while task is running
        nextRun = null;
    }
    else {
        if ((thisRunData.attemptNumber === 1) !== (thisRunData.firstAttempt === undefined)) {
            throw new AssertionError('Expect firstAttempt is defined when attemptNumber is larger than 1');
        }
        const options = __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_options, "f").onError;
        if (options === null) {
            nextRun = null;
        }
        else if (typeof options === 'function') {
            nextRun = options(caughtValue, {
                attemptNumber: thisRunData.attemptNumber,
                startTime,
                endTime,
                firstAttemptStartTime: (_b = (_a = thisRunData.firstAttempt) === null || _a === void 0 ? void 0 : _a.startTime) !== null && _b !== void 0 ? _b : startTime,
                firstAttemptEndTime: (_d = (_c = thisRunData.firstAttempt) === null || _c === void 0 ? void 0 : _c.endTime) !== null && _d !== void 0 ? _d : endTime
            }, __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_context, "f"));
        }
        else if (thisRunData.attemptNumber >= ((_e = options.attempt) !== null && _e !== void 0 ? _e : Infinity)) {
            nextRun = null;
        }
        else {
            nextRun = new Date(endTime.getTime() + options.delay);
        }
    }
    // Use next run time to set next run data
    if (nextRun === null) {
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_nextRunData, null, "f");
    }
    else {
        if (thisRunData === null) {
            throw new AssertionError('Expect thisRunData is not null');
        }
        if (typeof nextRun === 'number') {
            nextRun = new Date(Date.now() + nextRun);
        }
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_nextRunData, {
            startTime: nextRun,
            timer: setTimeout(__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_instances, "m", _SingleInstanceTaskScheduler_runTask).bind(this), nextRun.getTime() - Date.now()),
            attemptNumber: thisRunData.attemptNumber + 1
        }, "f");
        if (thisRunData.attemptNumber === 1) {
            __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f").firstAttempt = {
                startTime,
                endTime
            };
        }
        else {
            __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_nextRunData, "f").firstAttempt = thisRunData.firstAttempt;
        }
    }
}, _SingleInstanceTaskScheduler_runTask = function _SingleInstanceTaskScheduler_runTask() {
    if (__classPrivateFieldGet(this, _SingleInstanceTaskScheduler_taskRunningPromise, "f") !== null) {
        return;
    }
    // Ensure `taskRunningPromise` is not null, to produce correct `running` property
    __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_taskRunningPromise, true, "f");
    __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_taskRunningPromise, (async () => {
        let taskResult;
        const startTime = new Date();
        try {
            taskResult = {
                success: true,
                returnValue: await __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_task, "f").call(this, __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_context, "f"))
            };
        }
        catch (error) {
            taskResult = {
                success: false,
                caughtValue: error
            };
        }
        const endTime = new Date();
        if (!taskResult.success) {
            __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_instances, "m", _SingleInstanceTaskScheduler_scheduleWithErrorResult).call(this, taskResult.caughtValue, startTime, endTime);
            throw taskResult.caughtValue;
        }
        else {
            __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_instances, "m", _SingleInstanceTaskScheduler_scheduleWithSuccessResult).call(this, taskResult.returnValue, startTime, endTime);
        }
        return taskResult.returnValue;
    })(), "f");
    // Note: `finally()` is not executed immediately
    __classPrivateFieldGet(this, _SingleInstanceTaskScheduler_taskRunningPromise, "f").finally(() => {
        __classPrivateFieldSet(this, _SingleInstanceTaskScheduler_taskRunningPromise, null, "f");
        // Avoid unhandled rejection by catching
    }).catch(error => {
        if (error instanceof AssertionError) {
            throw error;
        }
    });
};
